import re
from enum import Enum
from typing import Callable, List, Optional, Iterable, Union, Literal, Any

from loguru import logger


class Language(str, Enum):
    """Enum of the programming languages."""

    CPP = "cpp"
    GO = "go"
    JAVA = "java"
    KOTLIN = "kotlin"
    JS = "js"
    TS = "ts"
    PHP = "php"
    PROTO = "proto"
    PYTHON = "python"
    RST = "rst"
    RUBY = "ruby"
    RUST = "rust"
    SCALA = "scala"
    SWIFT = "swift"
    MARKDOWN = "markdown"
    LATEX = "latex"
    HTML = "html"
    SOL = "sol"
    CSHARP = "csharp"
    COBOL = "cobol"
    C = "c"
    LUA = "lua"
    PERL = "perl"
    HASKELL = "haskell"
    ELIXIR = "elixir"


class RecursiveCharacterTextSplitter:
    def __init__(self,
            separators: Optional[List[str]] = None,
            keep_separator: Union[bool, Literal["start", "end"]] = False,
            is_separator_regex: bool = False,
            strip_whitespace: bool = True,
            chunk_size: int = 500,
            chunk_overlap: int = 100
    ):
        if chunk_overlap > chunk_size:
            raise ValueError(
                f"Got a larger chunk overlap ({chunk_overlap}) than chunk size "
                f"({chunk_size}), should be smaller."
            )
        self.chunk_size = chunk_size
        self.separators = separators or ["\n\n", "\n", " ", ""]
        self._is_separator_regex = is_separator_regex
        self._keep_separator = keep_separator  # Whether to keep the separator in the chunks
        self._length_function: Callable[[str], int] = len
        self._strip_whitespace = strip_whitespace  # If `True`, strips whitespace from the start and end of every document
        self._chunk_overlap = chunk_overlap  # Overlap in characters between chunks

    def split_text(self, text: str) -> List[str]:
        return self._split_text(text, self.separators)

    def _split_text(self, text: str, separators: List[str]) -> List[str]:
        final_chunks = []
        # Get appropriate separator to use
        separator = self.separators[-1]
        new_separators = []
        for i, _s in enumerate(self.separators):
            _separator = _s if self._is_separator_regex else re.escape(_s)
            if _s == "":
                separator = _s
                break
            if re.search(_separator, text):
                separator = _s
                new_separators = separators[i + 1:]
                break

        _separator = separator if self._is_separator_regex else re.escape(separator)
        splits = self._split_text_with_regex(text, _separator, self._keep_separator)

        # Now go merging things, recursively splitting longer texts.
        _good_splits = []
        _separator = "" if self._keep_separator else separator
        for s in splits:
            if self._length_function(s) < self.chunk_size:
                _good_splits.append(s)
            else:
                if _good_splits:
                    merged_text = self._merge_splits(_good_splits, _separator)
                    final_chunks.extend(merged_text)
                    _good_splits = []
                if not new_separators:
                    final_chunks.append(s)
                else:
                    other_info = self._split_text(s, new_separators)
                    final_chunks.extend(other_info)
        if _good_splits:
            merged_text = self._merge_splits(_good_splits, _separator)
            final_chunks.extend(merged_text)
        return final_chunks

    @staticmethod
    def _split_text_with_regex(text: str, separator: str, keep_separator: bool) -> List[str]:
        if separator:
            if keep_separator:
                _splits = re.split(f"({separator})", text)
                splits = [_splits[i] + _splits[i + 1] for i in range(1, len(_splits), 2)]
                if len(_splits) % 2 == 0:
                    splits += _splits[-1:]
                splits = [_splits[0]] + splits
            else:
                splits = re.split(separator, text)
        else:
            splits = list(text)
        return [s for s in splits if s != ""]

    def _join_docs(self, docs: List[str], separator: str) -> Optional[str]:
        text = separator.join(docs)
        if self._strip_whitespace:
            text = text.strip()
        if text == "":
            return None
        else:
            return text

    def _merge_splits(self, splits: Iterable[str], separator: str) -> List[str]:
        # We now want to combine these smaller pieces into medium size
        # chunks to send to the LLM.
        separator_len = self._length_function(separator)

        docs = []
        current_doc: List[str] = []
        total = 0
        for d in splits:
            _len = self._length_function(d)
            if (
                    total + _len + (separator_len if len(current_doc) > 0 else 0)
                    > self.chunk_size
            ):
                if total > self.chunk_size:
                    logger.warning(
                        f"Created a chunk of size {total}, "
                        f"which is longer than the specified {self.chunk_size}"
                    )
                if len(current_doc) > 0:
                    doc = self._join_docs(current_doc, separator)
                    if doc is not None:
                        docs.append(doc)
                    # Keep on popping if:
                    # - we have a larger chunk than in the chunk overlap
                    # - or if we still have any chunks and the length is long
                    while total > self._chunk_overlap or (
                            total + _len + (separator_len if len(current_doc) > 0 else 0)
                            > self.chunk_size
                            and total > 0
                    ):
                        total -= self._length_function(current_doc[0]) + (
                            separator_len if len(current_doc) > 1 else 0
                        )
                        current_doc = current_doc[1:]
            current_doc.append(d)
            total += _len + (separator_len if len(current_doc) > 1 else 0)
        doc = self._join_docs(current_doc, separator)
        if doc is not None:
            docs.append(doc)
        return docs

    @classmethod
    def from_language(
            cls, language: Language, **kwargs: Any
    ) -> "RecursiveCharacterTextSplitter":
        separators = cls.get_separators_for_language(language)
        return cls(separators=separators, is_separator_regex=True, **kwargs)

    @staticmethod
    def get_separators_for_language(language: Language) -> List[str]:
        if language == Language.CPP:
            return [
                # Split along class definitions
                "\nclass ",
                # Split along function definitions
                "\nvoid ",
                "\nint ",
                "\nfloat ",
                "\ndouble ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.GO:
            return [
                # Split along function definitions
                "\nfunc ",
                "\nvar ",
                "\nconst ",
                "\ntype ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.JAVA:
            return [
                # Split along class definitions
                "\nclass ",
                # Split along method definitions
                "\npublic ",
                "\nprotected ",
                "\nprivate ",
                "\nstatic ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.KOTLIN:
            return [
                # Split along class definitions
                "\nclass ",
                # Split along method definitions
                "\npublic ",
                "\nprotected ",
                "\nprivate ",
                "\ninternal ",
                "\ncompanion ",
                "\nfun ",
                "\nval ",
                "\nvar ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nwhen ",
                "\ncase ",
                "\nelse ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.JS:
            return [
                # Split along function definitions
                "\nfunction ",
                "\nconst ",
                "\nlet ",
                "\nvar ",
                "\nclass ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nswitch ",
                "\ncase ",
                "\ndefault ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.TS:
            return [
                "\nenum ",
                "\ninterface ",
                "\nnamespace ",
                "\ntype ",
                # Split along class definitions
                "\nclass ",
                # Split along function definitions
                "\nfunction ",
                "\nconst ",
                "\nlet ",
                "\nvar ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nswitch ",
                "\ncase ",
                "\ndefault ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.PHP:
            return [
                # Split along function definitions
                "\nfunction ",
                # Split along class definitions
                "\nclass ",
                # Split along control flow statements
                "\nif ",
                "\nforeach ",
                "\nwhile ",
                "\ndo ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.PROTO:
            return [
                # Split along message definitions
                "\nmessage ",
                # Split along service definitions
                "\nservice ",
                # Split along enum definitions
                "\nenum ",
                # Split along option definitions
                "\noption ",
                # Split along import statements
                "\nimport ",
                # Split along syntax declarations
                "\nsyntax ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.PYTHON:
            return [
                # First, try to split along class definitions
                "\nclass ",
                "\ndef ",
                "\n\tdef ",
                # Now split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.RST:
            return [
                # Split along section titles
                "\n=+\n",
                "\n-+\n",
                "\n\\*+\n",
                # Split along directive markers
                "\n\n.. *\n\n",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.RUBY:
            return [
                # Split along method definitions
                "\ndef ",
                "\nclass ",
                # Split along control flow statements
                "\nif ",
                "\nunless ",
                "\nwhile ",
                "\nfor ",
                "\ndo ",
                "\nbegin ",
                "\nrescue ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.ELIXIR:
            return [
                # Split along method function and module definiton
                "\ndef ",
                "\ndefp ",
                "\ndefmodule ",
                "\ndefprotocol ",
                "\ndefmacro ",
                "\ndefmacrop ",
                # Split along control flow statements
                "\nif ",
                "\nunless ",
                "\nwhile ",
                "\ncase ",
                "\ncond ",
                "\nwith ",
                "\nfor ",
                "\ndo ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.RUST:
            return [
                # Split along function definitions
                "\nfn ",
                "\nconst ",
                "\nlet ",
                # Split along control flow statements
                "\nif ",
                "\nwhile ",
                "\nfor ",
                "\nloop ",
                "\nmatch ",
                "\nconst ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.SCALA:
            return [
                # Split along class definitions
                "\nclass ",
                "\nobject ",
                # Split along method definitions
                "\ndef ",
                "\nval ",
                "\nvar ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nmatch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.SWIFT:
            return [
                # Split along function definitions
                "\nfunc ",
                # Split along class definitions
                "\nclass ",
                "\nstruct ",
                "\nenum ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\ndo ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.MARKDOWN:
            return [
                # First, try to split along Markdown headings (starting with level 2)
                "\n#{1,6} ",
                # Note the alternative syntax for headings (below) is not handled here
                # Heading level 2
                # ---------------
                # End of code block
                "```\n",
                # Horizontal lines
                "\n\\*\\*\\*+\n",
                "\n---+\n",
                "\n___+\n",
                # Note that this splitter doesn't handle horizontal lines defined
                # by *three or more* of ***, ---, or ___, but this is not handled
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.LATEX:
            return [
                # First, try to split along Latex sections
                "\n\\\\chapter{",
                "\n\\\\section{",
                "\n\\\\subsection{",
                "\n\\\\subsubsection{",
                # Now split by environments
                "\n\\\\begin{enumerate}",
                "\n\\\\begin{itemize}",
                "\n\\\\begin{description}",
                "\n\\\\begin{list}",
                "\n\\\\begin{quote}",
                "\n\\\\begin{quotation}",
                "\n\\\\begin{verse}",
                "\n\\\\begin{verbatim}",
                # Now split by math environments
                "\n\\\begin{align}",
                "$$",
                "$",
                # Now split by the normal type of lines
                " ",
                "",
            ]
        elif language == Language.HTML:
            return [
                # First, try to split along HTML tags
                "<body",
                "<div",
                "<p",
                "<br",
                "<li",
                "<h1",
                "<h2",
                "<h3",
                "<h4",
                "<h5",
                "<h6",
                "<span",
                "<table",
                "<tr",
                "<td",
                "<th",
                "<ul",
                "<ol",
                "<header",
                "<footer",
                "<nav",
                # Head
                "<head",
                "<style",
                "<script",
                "<meta",
                "<title",
                "",
            ]
        elif language == Language.CSHARP:
            return [
                "\ninterface ",
                "\nenum ",
                "\nimplements ",
                "\ndelegate ",
                "\nevent ",
                # Split along class definitions
                "\nclass ",
                "\nabstract ",
                # Split along method definitions
                "\npublic ",
                "\nprotected ",
                "\nprivate ",
                "\nstatic ",
                "\nreturn ",
                # Split along control flow statements
                "\nif ",
                "\ncontinue ",
                "\nfor ",
                "\nforeach ",
                "\nwhile ",
                "\nswitch ",
                "\nbreak ",
                "\ncase ",
                "\nelse ",
                # Split by exceptions
                "\ntry ",
                "\nthrow ",
                "\nfinally ",
                "\ncatch ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.SOL:
            return [
                # Split along compiler information definitions
                "\npragma ",
                "\nusing ",
                # Split along contract definitions
                "\ncontract ",
                "\ninterface ",
                "\nlibrary ",
                # Split along method definitions
                "\nconstructor ",
                "\ntype ",
                "\nfunction ",
                "\nevent ",
                "\nmodifier ",
                "\nerror ",
                "\nstruct ",
                "\nenum ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\ndo while ",
                "\nassembly ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.COBOL:
            return [
                # Split along divisions
                "\nIDENTIFICATION DIVISION.",
                "\nENVIRONMENT DIVISION.",
                "\nDATA DIVISION.",
                "\nPROCEDURE DIVISION.",
                # Split along sections within DATA DIVISION
                "\nWORKING-STORAGE SECTION.",
                "\nLINKAGE SECTION.",
                "\nFILE SECTION.",
                # Split along sections within PROCEDURE DIVISION
                "\nINPUT-OUTPUT SECTION.",
                # Split along paragraphs and common statements
                "\nOPEN ",
                "\nCLOSE ",
                "\nREAD ",
                "\nWRITE ",
                "\nIF ",
                "\nELSE ",
                "\nMOVE ",
                "\nPERFORM ",
                "\nUNTIL ",
                "\nVARYING ",
                "\nACCEPT ",
                "\nDISPLAY ",
                "\nSTOP RUN.",
                # Split by the normal type of lines
                "\n",
                " ",
                "",
            ]
        elif language == Language.LUA:
            return [
                # Split along variable and table definitions
                "\nlocal ",
                # Split along function definitions
                "\nfunction ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nrepeat ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.HASKELL:
            return [
                # Split along function definitions
                "\nmain :: ",
                "\nmain = ",
                "\nlet ",
                "\nin ",
                "\ndo ",
                "\nwhere ",
                "\n:: ",
                "\n= ",
                # Split along type declarations
                "\ndata ",
                "\nnewtype ",
                "\ntype ",
                "\n:: ",
                # Split along module declarations
                "\nmodule ",
                # Split along import statements
                "\nimport ",
                "\nqualified ",
                "\nimport qualified ",
                # Split along typeclass declarations
                "\nclass ",
                "\ninstance ",
                # Split along case expressions
                "\ncase ",
                # Split along guards in function definitions
                "\n| ",
                # Split along record field declarations
                "\ndata ",
                "\n= {",
                "\n, ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language in Language._value2member_map_:
            raise ValueError(f"Language {language} is not implemented yet!")
        else:
            raise ValueError(
                f"Language {language} is not supported! "
                f"Please choose from {list(Language)}"
            )


if __name__ == '__main__':
    character_text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=512,
        chunk_overlap=0,
        separators=["\n\n", "\n", "。", "．", ".", "；", ";", "，", ",", " ", ""]
    )
    text = """
    三、技术指标要求
1.功能指标要求
1）Z区方向航空影像大数据分析应用
（1）航空影像目标标注功能。
（1.1）航空影像数据管理：支持航空影像数据集构建、查询、修改、删除等功能。
★（1.2）航空影像人工标注：支持对已搜集到的低清晰度实侦高价值航空影像数据进行人工目标打标处理，标注航空影像目标类别、目标具体信息（包括但不限于目标大小、目标国别、目标作Z能力等）、目标位置等具体信息。
（1.3）航空影像自动辅助标注：能够提供自动化批量航空影像数据标注推荐，自动输出目标类别等可自动化生成的信息，并支持对标注进行人工补充修正。
（1.4）基于特性模型的可见光、红外、SAR航空影像生成：支持对可见光图像、红外、SAR图像进行转换、显示、特征提取等操作，并实现基于特征模型的图像编辑、生成、模糊仿真、目标变换、目标分割等功能。
（1.5）基于生成对抗模型的可见光、红外、SAR航空影像生成：支持可见光、红外、SAR航空影像对抗模型的构建和对抗图像生成，可实现图像风格的转换和实测图像与生成图像之间的相似度度量。
2）航空影像目标识别模型训练功能。
（2.1）航空影像目标识别训练数据集管理：支持可见光、红外和SAR多传感器数据集的创建、下载、合并、查询等一系列功能，并可以对数据标签进行查看、编辑、统计等操作。
★（2.2）基于单阶段架构的高效识别检测算法：能够根据已标注的高质量航空影像训练数据，基于任务需求和计算资源加载合适规模的单阶段架构的识别检测算法，进行高效率模型训练，优化识别模型。支持自定义任务管理、训练算法配置、模型训练、训练过程监控、断点保存和恢复、模型测试、模型查询等功能。
★（2.3）基于双阶段架构的高精度识别检测算法：能够根据已标注的高质量航空影像训练数据，基于任务需求和计算资源加载合适规模的双阶段架构的识别检测算法，进行高精度模型训练，优化识别模型。支持自定义任务管理、训练算法配置、模型训练、训练过程监控、断点保存和恢复、模型测试、模型查询等功能。
（2.4）航空影像识别效能分析：能够对可见光、红外、SAR航空影像目标识别模型进行效能分析，支持训练和验证集划分、检测结果可视化、精确率和召回率计算、混淆矩阵计算和显示等功能。
（2.5）航空影像目标识别模型训练前端：支持使用图形化、交互式操作界面进行神经网络模型训练，可通过前端界面配置数据集、模型类别、模型结构、训练超参数等信息，实现神经网络模型的训练，从而降低培训成本。
3）航空影像目标识别功能。
★（3.1）航空影像目标检测。能够通过训练好的识别模型，实现单幅影像及批量影像的目标检测，以目标框标识识别出的飞行器目标，并输出目标属于飞行器的概率值。
（3.2）航空影像目标分类。能够通过训练好的分类模型，在目标检测结果的基础上，实现飞行器目标的分类，支撑包括目标类别、具体型号等信息的识别，辅助快速生成高价值目标QB产品。
（3.3）航空影像分析结果数据管理：能够基于识别结果，对各类目标进行分类归档管理，支持对结果的查看、查询、修正、导出等功能。
4）航空影像比对与管理功能。
★（4.1）航空影像目标情报影像比对。能够对航空影像历史数据和实测回传数据进行显示、查询、修改、删除等操作。能够基于航空影像目标识别结果，对重要地点不同时间点的航空影像进行比对分析，识别区域中目标的变化情况。
（4.2）航空影像目标时间线管理。能够对航空影像历史数据进行基于日期的检索、排序、修改、删除。能够基于航空影像目标识别结果对有价值目标或区域进行时间线管理，即分析连续时间片上的动态变化情况。
（4.3）航空影像目标情报跟踪。能够实现航空影像中高价值目标的情报跟踪，对目标的动态进行追踪，辅助实现对特定目标情报的持续跟踪分析与研判处置。
    """
    result = character_text_splitter.split_text(text)
    print(result)



import os
from abc import ABC, abstractmethod
from typing import List, Dict

from pydantic import BaseModel, ConfigDict

from pfip.base.errors import UnImplementException
from pfip.base.parser.parser_model import ParseResult, TitleNode
from pfip.base.util.file_util import get_temp_dir, clean_file_name


class Parser(ABC, BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    empty_line_pattern: str = r'^\s*$'
    full_title_max_len: int = 200

    @abstractmethod
    def support(self, file_ext: str) -> bool:
        return False

    @abstractmethod
    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        raise UnImplementException("未实现...")

    @staticmethod
    def is_empty_line(line) -> bool:
        return len(line.strip()) == 0

    def fill_titles(self, titles: List[TitleNode]):
        """
            为 sn|pid 进行赋值
        """
        for idx, t in enumerate(titles):
            t.sn = idx + 1
        self.fill_title_pid(titles)
        self.fill_full_title(titles)

    @staticmethod
    def fill_title_pid(titles: List[TitleNode]):
        """
        根据title的前后关系以及level_type来进行确定
        """
        stack = []
        for i, title_node in enumerate(titles):
            while stack and stack[-1].title_level >= title_node.title_level:
                stack.pop()
            if stack:
                title_node.pid = stack[-1].id
            stack.append(title_node)

    def fill_full_title(self, nodes: List[TitleNode]):
        node_dict: Dict[str, TitleNode] = {node.id: node for node in nodes}
        for node in nodes:
            current = node
            full_title_list = []
            while current.pid is not None:
                parent = node_dict.get(current.pid)
                if parent:
                    full_title_list.insert(0, parent.title)
                    current = parent
            node.full_title = " - ".join(full_title_list + [node.title])
            if len(node.full_title) > self.full_title_max_len:
                # 如果full_title超过限制,则只是用首个节点的title
                node.full_title = nodes[0].title

    @staticmethod
    def get_home_dir(file_path: str):
        file_name_with_ext = os.path.basename(file_path)
        file_name, file_ext = os.path.splitext(file_name_with_ext)
        file_name = clean_file_name(file_name)
        file_ext = file_ext.replace(".", "")
        home_dir = os.path.join(get_temp_dir(), file_ext, file_name)
        if not os.path.exists(home_dir):
            os.makedirs(home_dir)
        return home_dir



import threading
import time


class SnowflakeIdWorker:
    twepoch = 1288834974657  # 起始时间戳，这里假设为2010-11-04 09:42:54.657 UTC
    worker_id_bits = 5  # 机器id所占的位数
    datacenter_id_bits = 5  # 数据中心id所占的位数
    max_worker_id = -1 ^ (-1 << worker_id_bits)  # 可用的最大机器id，结果是31
    max_datacenter_id = -1 ^ (-1 << datacenter_id_bits)  # 可用的最大数据中心id，结果是31
    sequence_bits = 12  # 序列在id中占的位数
    worker_id_shift = sequence_bits  # 机器id向左移12位
    datacenter_id_shift = sequence_bits + worker_id_bits  # 数据中心id向左移17位(12+5)
    timestamp_left_shift = sequence_bits + worker_id_bits + datacenter_id_bits  # 时间截向左移22位(5+5+12)
    sequence_mask = -1 ^ (-1 << sequence_bits)  # 生成序列的掩码，这里为4095
    last_timestamp = -1  # 上次生成ID的时间截
    # 锁对象
    lock = threading.Lock()

    def __init__(self, datacenter_id, worker_id, sequence=0):
        if worker_id > self.max_worker_id or worker_id < 0:
            raise ValueError('worker_id值越界')
        if datacenter_id > self.max_datacenter_id or datacenter_id < 0:
            raise ValueError('datacenter_id值越界')
        self.datacenter_id = datacenter_id
        self.worker_id = worker_id
        self.sequence = sequence

    def _til_next_millis(self, last_timestamp):
        timestamp = int(time.time() * 1000)
        while timestamp <= last_timestamp:
            timestamp = int(time.time() * 1000)
        return timestamp

    def next_id(self):
        with self.lock:
            timestamp = int(time.time() * 1000)

            if self.last_timestamp == timestamp:
                self.sequence = (self.sequence + 1) & self.sequence_mask
                if self.sequence == 0:
                    timestamp = self._til_next_millis(self.last_timestamp)
            else:
                self.sequence = 0

            self.last_timestamp = timestamp

            return ((timestamp - self.twepoch) << self.timestamp_left_shift) | \
                (self.datacenter_id << self.datacenter_id_shift) | \
                (self.worker_id << self.worker_id_shift) | \
                self.sequence


id_generator = SnowflakeIdWorker(datacenter_id=10, worker_id=10)

if __name__ == '__main__':
    for _ in range(10):
        print(id_generator.next_id())



from typing import List

import fitz

from pfip.base.parser.parser_model import TitleNode, AtomItem
from pfip.pdf.title.base import TitleExtractor


class BookMarkTitleExtractor(TitleExtractor):
    def __call__(self, file_path: str, items: List[AtomItem] = []) -> List[TitleNode]:
        titles = []
        with fitz.open(file_path) as pdf:
            for index, entry in enumerate(pdf.get_toc()):
                item = TitleNode(
                    title_level=entry[0],
                    title=entry[1],
                    page_number=entry[2],
                    sn=index + 1
                )
                titles.append(item)
        return titles



from typing import List, Iterable

from panshi2task.paddle_client import BoxRecItem
from pydantic import BaseModel

from pfip.base.constant import UNKOWN_TITLE_LEVEL
from pfip.base.parser.parser_model import AtomItem, TitleAtomItem, ImageAtomItem, \
    ImageCaptionAtomItem, TextAtomItem, ScannedPDFTableAtomItem


class ScannedPdfElementHandler(BaseModel):
    def run(self, itr: Iterable[BoxRecItem], file_path: str) -> List[AtomItem]:
        rtn = []
        last_page_num = 0
        for rec_item in itr:
            cur_page_number = rec_item.page_num
            if rec_item.type == "title":
                rtn.extend(self.parse_title(rec_item))
            elif rec_item.type == "table":
                rtn.extend(self.parse_table(rec_item))
            elif rec_item.type == "figure":
                rtn.extend(self.parse_image(rec_item))
            elif rec_item.type in ["text", "reference"]:
                rtn.extend(self.parse_text_or_reference(rec_item))
            elif rec_item.type == "figure_caption":
                rtn.extend(self.parse_image_caption(rec_item))
            else:
                continue
            if last_page_num != cur_page_number:
                last_page_num = cur_page_number
        return rtn

    @staticmethod
    def parse_title(rec_item: BoxRecItem) -> List[TitleAtomItem]:
        rtn = []
        if not rec_item.res:
            return rtn
        title: str = ""
        for item in rec_item.res:
            if not item.text:
                continue
            title += item.text
        rtn.append(TitleAtomItem.instance(title, UNKOWN_TITLE_LEVEL, rec_item.page_num))
        return rtn

    @staticmethod
    def parse_text_or_reference(rec_item: BoxRecItem) -> List[TextAtomItem]:
        rtn = []
        if not rec_item.res:
            return rtn

        for item in rec_item.res:
            if not item.text:
                continue
            rtn.append(TextAtomItem.instance(item.text, rec_item.page_num))
        return rtn

    @staticmethod
    def parse_table(rec_item: BoxRecItem) -> List[ScannedPDFTableAtomItem]:
        rtn = []
        if not rec_item.res:
            return rtn

        for item in rec_item.res:
            if not item.table_html:
                continue
            rtn.append(ScannedPDFTableAtomItem.simple(item.table_html, page_num=rec_item.page_num))
        return rtn

    @staticmethod
    def parse_image(rec_item: BoxRecItem) -> List[ImageAtomItem]:
        rtn = []
        if rec_item.img_url:
            rtn.append(ImageAtomItem.simple(rec_item.img_url, "", rec_item.page_num))
        return rtn

    @staticmethod
    def parse_image_caption(rec_item: BoxRecItem) -> List[ImageCaptionAtomItem]:
        if not rec_item.res:
            return []
        image_desc = []
        for item in rec_item.res:
            if not item.text:
                continue
            image_desc.append(item.text)
        image_desc_str = "".join(image_desc)
        return [ImageCaptionAtomItem.simple(image_desc_str, rec_item.page_num)]



import os.path
from typing import Dict, Optional

import PyPDF2
from loguru import logger
from pydantic import Field, model_validator
from typing_extensions import Self

from pfip.base.constant import TFileExt, TPdf
from pfip.base.parser.base import Parser
from pfip.base.parser.parser_model import ParseResult
from pfip.base.util.file_util import get_file_name_without_ext
from pfip.pdf.analyse.pdf_analyser import PdfAnalyser
from pfip.pdf.scanned.scanned_pdf_parser import ScannedPdfParser
from pfip.pdf.text.text_pdf_parser import TextPdfParser


class ProxyPdfParser(Parser):
    embedding_server_url: Optional[str] = None
    paddle_server_url: Optional[str] = None
    handlers: Dict[TPdf, Parser] = Field(default={}, exclude=True)
    pdf_analyser: PdfAnalyser = Field(default=PdfAnalyser(), exclude=True)

    @model_validator(mode='after')
    def init(self) -> Self:
        text_pdf_parser = TextPdfParser(
            embedding_server_url=self.embedding_server_url,
            paddle_server_url=self.paddle_server_url
        )
        self.handlers[TPdf.TEXT] = text_pdf_parser
        if self.paddle_server_url:
            scanned_pdf_parser = ScannedPdfParser(
                paddle_server_url=self.paddle_server_url,
                embedding_server_url=self.embedding_server_url
            )
            self.handlers[TPdf.SCANNED] = scanned_pdf_parser
        else:
            logger.warning("paddle_server_url未设置,扫描版PDF处理器将不可用...")
        return self

    def support(self, file_ext: str) -> bool:
        return file_ext.lower() == TFileExt.PDF

    def repair_pdf(self, file_path: str) -> str:
        home_dir = self.get_home_dir(file_path)
        with open(file_path, "rb") as file:
            reader = PyPDF2.PdfReader(file)
            basename = get_file_name_without_ext(file_path)
            repaired_path = os.path.join(home_dir, basename + "_repaired.pdf")
            with open(repaired_path, "wb") as new_file:
                writer = PyPDF2.PdfWriter()
                for i in range(len(reader.pages)):
                    writer.add_page(reader.pages[i])
                writer.write(new_file)
            return repaired_path

    def __call__(self, file_path: str, **kwargs) -> ParseResult:
        try:
            analyse_result = self.pdf_analyser(file_path)
        except Exception as e:
            logger.error(f"{e}, {file_path} 打开失败，修复该文档格式不规范问题")
            file_path = self.repair_pdf(file_path)
            analyse_result = self.pdf_analyser(file_path)

        pdf_type = analyse_result.pdf_type
        assert pdf_type in self.handlers, f"不支持的处理类型:{pdf_type}"
        logger.info("{}==>pdf类型为{}", file_path, pdf_type)
        handler = self.handlers.get(pdf_type)
        return handler(file_path, analyse_result=analyse_result)



from abc import ABC, abstractmethod
from typing import List

from pydantic import BaseModel, ConfigDict

from pfip.base.errors import UnImplementException
from pfip.base.parser.parser_model import ParseResult
from pfip.base.splitter.splitter_model import Chunk, Sentence


class ChunkSplitter(ABC, BaseModel):
    """chunk切分处理器"""
    model_config = ConfigDict(arbitrary_types_allowed=True)
    max_chunk_size: int = 2000

    @abstractmethod
    def support(self, file_ext: str) -> bool:
        return False

    @abstractmethod
    def __call__(self, parse_result: ParseResult) -> List[Chunk]:
        raise UnImplementException("未实现...")


class SentenceSplitter(ABC, BaseModel):
    """sentence切分处理器"""
    model_config = ConfigDict(arbitrary_types_allowed=True)
    max_sentence_size: int = 80

    @abstractmethod
    def support(self, file_ext: str) -> bool:
        return False

    @abstractmethod
    def __call__(self, chunk: Chunk) -> List[Sentence]:
        raise UnImplementException("未实现...")



from typing import List

from pydantic import BaseModel

from pfip.base.constant import TAtomItem
from pfip.base.parser.parser_model import AtomItem


class PPTRePair(BaseModel):


    @staticmethod
    def run(items: List[AtomItem]) -> List[AtomItem]:
        title_items_with_idx = [(idx, item) for idx, item in enumerate(items) if item.item_type == TAtomItem.TITLE]
        for idx, title in title_items_with_idx:
            if idx + 1 >= len(items):
                continue
            next_item = items[idx + 1]
            if title.content == next_item.content:
                items[idx + 1] = None
        return [item for item in items if item]



import fitz
from pydantic import BaseModel

from pfip.base.constant import TPdf


class PdfTypeAnalyser(BaseModel):
    max_extract_page_num: int = 20
    """采样的页码数量"""
    scanned_pdf_threshold: float = 0.7

    def run(self, pdf_path: str) -> TPdf:
        """
            如果图片页所占比重超过threshold,则认为是扫描件,反之为文字版
        """
        with fitz.open(pdf_path) as doc:
            total_page_num = doc.page_count
            assert total_page_num != 0
            min_page_num = min(total_page_num, self.max_extract_page_num)
            cur_page = 0
            image_page_num = 0
            for page_num in range(min_page_num):
                p = doc.load_page(page_num)
                if cur_page >= min_page_num:
                    break
                if not p.get_text("text"):
                    image_page_num += 1
                cur_page += 1

            radio = image_page_num / min_page_num
            if radio > self.scanned_pdf_threshold:
                return TPdf.SCANNED
            else:
                return TPdf.TEXT



import math
from typing import List, Optional

from loguru import logger
from panshi2task.embedding_client import EmbeddingTaskClient
from pydantic import BaseModel, ConfigDict, model_validator
from typing_extensions import Self

from pfip.base.constant import TAtomItem
from pfip.base.parser.parser_model import TextAtomItem, TitleNode
from pfip.base.util.common import min_edit_distance


class TitleConnector(BaseModel):
    """
            便利title,从item中找寻最相似的候选集. 然后再通过语义相似度比较找到最终的1个.
            如果没找到: 则跳过即可
            如果找到: 则修改
             下次找的位置变换 :
                从
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    embedding_server_url: Optional[str] = None
    client: Optional[EmbeddingTaskClient] = None
    max_ld_radio: float = 0.5
    """最小编辑距离比率=编辑距离/max(字符串1长度+字符串2长度)"""
    min_score: float = 0.8

    @model_validator(mode='after')
    def init(self) -> Self:
        if self.embedding_server_url:
            self.client = EmbeddingTaskClient(self.embedding_server_url)
        else:
            logger.warning("embedding_server_url未设置,语义相似度功能将不可用...")
        return self


    def calculate_str2_max_length(self, str1_len: int) -> int:
        """
        在已知str1的长度以及比率时,得到str2的最大可能长度
        那么存在: len(str2) -5 / len(str1)+len(str2) = 0.3  也就是: len(str2) = 1.3*len(str1)/0.7
        """
        c1 = 1 + self.max_ld_radio
        c2 = 1 - self.max_ld_radio
        return math.ceil(c1 * str1_len / c2)


    def find_similar_title(self, title: TitleNode, candcates: List[tuple[TextAtomItem, int]]) -> int:
        """
            step1: 现根据title长度以及比率,获取TextAtomItem的最大可能长度,根据此长度过滤掉部分数据
            step2: 根据最小编辑距离,得到一批候选集
            step3: 根据语义相似度,返回最相似的一个
        """
        failed_result = -1
        item_max_len = self.calculate_str2_max_length(len(title.title))
        candcates1 = [item for item in candcates if item[0].item_type in [TAtomItem.TITLE, TAtomItem.TEXT]]
        candcates2 = [item for item in candcates1 if len(item[0].content) <= item_max_len]
        candcates3 = []
        for item, idx in candcates2:
            max_len = max(len(title.title), len(item.content))
            radio = min_edit_distance(title.title, item.content) / max_len
            if radio <= self.max_ld_radio:
                candcates3.append((item, idx, radio))

        if not candcates3:
            return failed_result
        if self.client:
            texts = [c[0].content for c in candcates3]
            most_similar_text, score = self.client.text_similar(title.title, texts)
            if score > self.min_score:
                return [c[1] for c in candcates3 if c[0].content == most_similar_text][0]
            else:
                return failed_result
        else:
            min_tuple = min(candcates3, key=lambda tuple_item: tuple_item[2])
            return min_tuple[1]



import re
from typing import List, Tuple

from pydantic import BaseModel

from pfip.base.constant import TAtomItem
from pfip.base.parser.parser_model import AtomItem, TextAtomItem
from pfip.base.util.str_util import clean_str


class ImageDescRepair(BaseModel):
    pattern: str = r"^图\d+.{1,40}$"

    def _find_image_desc_item(self, idx: int, items: List[AtomItem]) -> Tuple[int, AtomItem | None]:
        if idx + 1 < len(items):
            right = items[idx + 1]
            content = clean_str(right.content)
            if isinstance(right, TextAtomItem) and re.match(self.pattern, content):
                return idx + 1, right
        if idx - 1 >= 0:
            left = items[idx - 1]
            content = clean_str(left.content)
            if isinstance(left, TextAtomItem) and re.match(self.pattern, content):
                return idx - 1, left
        return -1, None

    def run(self, items: List[AtomItem]) -> List[AtomItem]:
        image_item_idxs = []
        need_remove_item_idxs = []
        for idx, item in enumerate(items):
            if item.item_type == TAtomItem.IMAGE:
                image_item_idxs.append(idx)
        for idx in image_item_idxs:
            image_desc_idx, image_desc_item = self._find_image_desc_item(idx, items)
            if image_desc_item:
                items[idx].content = image_desc_item.content
                need_remove_item_idxs.append(image_desc_idx)
        filtered_items = [item for idx, item in enumerate(items) if idx not in need_remove_item_idxs]
        return filtered_items


class TableDescRepair(BaseModel):
    pattern: str = r"^表\d+.{1,40}$"

    def _find_table_desc_item(self, idx: int, items: List[AtomItem]) -> Tuple[int, AtomItem | None]:
        if idx + 1 < len(items):
            right = items[idx + 1]
            content = clean_str(right.content)
            if isinstance(right, TextAtomItem) and re.match(self.pattern, content):
                return idx + 1, right
        if idx - 1 >= 0:
            left = items[idx - 1]
            content = clean_str(left.content)
            if isinstance(left, TextAtomItem) and re.match(self.pattern, content):
                return idx - 1, left
        return -1, None

    def run(self, items: List[AtomItem]) -> List[AtomItem]:
        table_item_idxs = []
        need_remove_item_idxs = []
        for idx, item in enumerate(items):
            if item.item_type == TAtomItem.TABLE:
                table_item_idxs.append(idx)
        for idx in table_item_idxs:
            table_desc_idx, table_desc_item = self._find_table_desc_item(idx, items)
            if table_desc_item:
                items[idx].content = table_desc_item.content
                need_remove_item_idxs.append(table_desc_idx)
        filtered_items = [item for idx, item in enumerate(items) if idx not in need_remove_item_idxs]
        return filtered_items



import itertools
from collections import defaultdict, Counter
from typing import List, Dict

import pandas as pd
from pandas import DataFrame
from pydantic import BaseModel, model_validator
from typing_extensions import Self

from pfip.base.constant import TAtomItem
from pfip.base.parser.parser_model import AtomItem, TableAtomItem, ImageAtomItem, TitleNode, TitleAtomItem, \
    TextPDFTableAtomItem
from pfip.base.parser.repair import TableDescRepair, ImageDescRepair
from pfip.base.util.rectangle import RectangleHelper
from pfip.pdf.title.title_item_connector import TitleConnector


class CrossPageTableRepair(BaseModel):
    """
        跨页表格发现及合并处理
    """

    @staticmethod
    def _merge_cross_page_tables(merge_table_idxs: List[int], items: List[AtomItem]):
        """
        合并表格
        :param merge_table_idxs:
        :param items:
        :return:
        """
        first_part_table: TextPDFTableAtomItem = items[merge_table_idxs[0]]
        first_header = first_part_table.df.columns.tolist()
        first_rows = first_part_table.df.values.tolist()
        replace_items = first_part_table.replace_items
        end_page = items[merge_table_idxs[-1]].end_page
        for idx in merge_table_idxs[1:]:
            header = items[idx].df.columns.tolist()
            rows = items[idx].df.values.tolist()
            if header != first_header:
                first_rows.append(header)
            first_rows.extend(rows)
            replace_items.extend(items[idx].replace_items)
        df = pd.DataFrame(first_rows, columns=first_header)
        items[merge_table_idxs[0]].df = df
        items[merge_table_idxs[0]].end_page = end_page

    @staticmethod
    def _is_top_start_table(table_item: AtomItem) -> bool:
        """
        判断表格是否是当前页顶部起始的表格
        :param table_item:
        :return:
        """
        if table_item.location.y1 / table_item.page_height > 0.9:
            return True
        return False

    @staticmethod
    def _is_adjoin_page_table(pre_table_item: AtomItem, current_table_item: AtomItem) -> bool:
        """
        判断是否是相邻页的两个表格
        :param pre_table_item:
        :param current_table_item:
        :return:
        """
        if pre_table_item.end_page + 1 == current_table_item.start_page:
            return True
        return False

    @staticmethod
    def _is_same_width_table(pre_table_item: AtomItem, current_table_item: AtomItem) -> bool:
        """
        判断表格宽度是否相同
        宽度误差在5以内都算相同
        :param pre_table_item:
        :param current_table_item:
        :return:
        """
        pre_table_width = pre_table_item.location.x1 - pre_table_item.location.x0
        current_table_width = current_table_item.location.x1 - current_table_item.location.x0
        return abs(pre_table_width - current_table_width) < 5

    def _is_cross_page_table(self, pre_table_item: AtomItem, current_table_item: AtomItem) -> bool:
        """
        判断前一个表格和当前表格是否是跨页的
        判断逻辑：
        1.当相邻两个表格header不同，但列数相同、表宽相同、页码相邻、header位置在页面顶部时，满足这些条件表示当前表格与上一个表格是跨页表格。
        2.相邻两个表格header相同、位置在页面顶部时,也是跨页表格。
        :param pre_table_item:
        :param current_table_item:
        :return:
        """
        pre_table_headers = pre_table_item.df.columns.tolist()
        current_table_headers = current_table_item.df.columns.tolist()
        # 是否是相同列数
        is_same_columns_count = len(pre_table_headers) == len(current_table_headers)
        # 是否是相邻页
        is_adjoin_page = self._is_adjoin_page_table(pre_table_item, current_table_item)
        # 是否相同宽度
        is_same_width = self._is_same_width_table(pre_table_item, current_table_item)
        # 是否是顶部起始
        is_top = self._is_top_start_table(current_table_item)
        if pre_table_headers == current_table_headers:
            if is_top and is_adjoin_page:
                return True
        else:
            if is_same_columns_count and is_adjoin_page and is_same_width:
                return True
        return False

    def run(self, items: List[AtomItem]) -> List[AtomItem]:
        """
        跨页表格处理
        逻辑：
          1.大于1个表格数才会考虑跨页合并问题
          2.先判断当前表格与前一个表格是否是跨页的,如果是，将索引位置添加到列表中，列表第一个元素为表格开始部分，剩余为待合并部分。
          3.合并完，将待合并部分的元素从列表中移除。
        :param items:
        :return:
        """
        table_item_idxs = []
        for idx, item in enumerate(items):
            if item.item_type == TAtomItem.TABLE:
                table_item_idxs.append(idx)
        if len(table_item_idxs) > 1:
            merge_table_idxs = []
            removed_idx = []
            for i, idx in enumerate(table_item_idxs[1:], start=1):
                if self._is_cross_page_table(items[table_item_idxs[i - 1]], items[idx]):
                    if merge_table_idxs:
                        merge_table_idxs.append(idx)
                        removed_idx.append(idx)
                    else:
                        merge_table_idxs = [table_item_idxs[i - 1], idx]
                        removed_idx.append(idx)
                else:
                    if merge_table_idxs:
                        self._merge_cross_page_tables(merge_table_idxs, items)
                        merge_table_idxs = []
            items = [item for index, item in enumerate(items) if index not in removed_idx]
        return items


class TableRectAngleContainer(BaseModel):
    table_idx: int
    table_item: TextPDFTableAtomItem
    indexs: List[int] = []
    rectangle_helper: RectangleHelper = None
    """连续的值"""

    @model_validator(mode="after")
    def init(self) -> Self:
        location = self.table_item.location
        self.rectangle_helper = RectangleHelper(
            bottom_left=(location.x0, location.y0), top_right=(location.x1, location.y1)
        )

    def get_sort_val(self) -> int:
        if self.indexs:
            return self.indexs[0]
        else:
            return 10000

    def is_empty(self) -> bool:
        return len(self.indexs) == 0

    def is_table_contained(self, idx: int, item: AtomItem) -> bool:
        x1, y1 = item.location.x0, item.location.y0
        x2, y2 = item.location.x1, item.location.y1

        self.rectangle_helper.contains((x1, y1), (x2, y2))
        if self.rectangle_helper.contains((x1, y1), (x2, y2)):
            self.indexs.append(idx)
            return True
        else:
            return False


class ImagePosRepair(BaseModel):
    @staticmethod
    def change_img_pos(items: List[AtomItem]) -> List[AtomItem]:
        not_image_items = [(idx, item) for idx, item in enumerate(items) if item.item_type != TAtomItem.IMAGE]
        image_items = [(idx, item) for idx, item in enumerate(items) if item.item_type == TAtomItem.IMAGE]
        changed_images_pos = []
        changed_images_ori_index = []
        for image_idx, image in image_items:
            l0 = image.location
            helper = RectangleHelper(bottom_left=(l0.x0, l0.y0), top_right=(l0.x1, l0.y1))
            for item_with_idx1, item_with_idx2 in zip(not_image_items[:-1], not_image_items[1:]):
                idx1, item1 = item_with_idx1[0], item_with_idx1[1]
                idx2, item2 = item_with_idx2[0], item_with_idx2[1]
                l1 = item1.location
                l2 = item2.location
                flag1 = not helper.in_top((l1.x0, l1.y0), (l1.x1, l1.y1))
                flag2 = helper.in_top((l2.x0, l2.y0), (l2.x1, l2.y1))
                if flag1 and flag2:
                    changed_images_pos.append(idx1)
                    changed_images_ori_index.append(image_idx)
                    break
        rtn = []
        for idx, item in enumerate(items):
            if idx in changed_images_pos:
                rtn.append(item)
                image_item = image_items[changed_images_pos.index(idx)]
                rtn.append(image_item[1])
            elif idx in changed_images_ori_index:
                continue
            else:
                rtn.append(item)
        return rtn

    def is_whole_page_image(self, image: ImageAtomItem) -> bool:
        """
        判断是否是整页的图片
        x0、y0 与 0 的误差在5之内
        :param image:
        :return:
        """
        if abs(image.location.x0 - 0) < 5 and abs(image.location.y0 - 0) < 5:
            return True
        return False

    def get_watermark_indexs(self, images: List[tuple[int, ImageAtomItem]]) -> List[int]:
        """
            localtion完全一致image>3,则被认为是水印
            额外再判断是否是整页的location，如果是，排除掉水印的可能
        """
        location_counts = defaultdict(int)
        watermark_indexs = []
        water_mark_image_locations = []
        for _, image in images:
            if not self.is_whole_page_image(image):
                location_key = (image.location.x0, image.location.x1, image.location.y0, image.location.y1)
                location_counts[location_key] += 1
        for location, count in location_counts.items():
            if count > 2:
                water_mark_image_locations.append(location)
        for idx, image in images:
            location_key = (image.location.x0, image.location.x1, image.location.y0, image.location.y1)
            if location_key in water_mark_image_locations:
                watermark_indexs.append(idx)
        return watermark_indexs

    def run(self, items: List[AtomItem]) -> List[AtomItem]:
        rtn = []
        images_with_idx: List[tuple[int, ImageAtomItem]] = []
        for idx, item in enumerate(items):
            if item.item_type == TAtomItem.IMAGE:
                images_with_idx.append((idx, item))
        # 去除水印
        watermark_indexs = self.get_watermark_indexs(images_with_idx)
        remove_watermark_items = [item for idx, item in enumerate(items) if idx not in watermark_indexs]
        # 改变图片坐标位置
        page_grouped_items = []
        for _, group in itertools.groupby(remove_watermark_items, key=lambda x: x.start_page):
            page_grouped_items.append(list(group))
        for group in page_grouped_items:
            image_dealed_group = self.change_img_pos(group)
            rtn.extend(image_dealed_group)
        return rtn


class TablePosRepair(BaseModel):
    """
        调整表格图片的位置,插入到正确的Item前后; 并移除表格图片对应的TextAtomItem
    """

    @staticmethod
    def deal_table(page_items: List[AtomItem]) -> List[AtomItem]:
        not_table_items = [(idx, item) for idx, item in enumerate(page_items) if item.item_type != TAtomItem.TABLE]
        table_items = [(idx, item) for idx, item in enumerate(page_items) if item.item_type == TAtomItem.TABLE]
        table_containers = []
        for table_idx, table in table_items:
            container = TableRectAngleContainer(table_idx=table_idx, table_item=table)
            for idx, item in not_table_items:
                in_container = container.is_table_contained(idx, item)
                if not in_container and not container.is_empty():
                    """当container中已经有了元素,并且遇到了终止的元素,意味着表格已结束"""
                    table_containers.append(container)
                    container = None
                    break
            if container:
                table_containers.append(container)
        sorted_table_containers = sorted(table_containers, key=lambda x: x.get_sort_val())
        for table_container in sorted_table_containers:
            if table_container.indexs:
                table_item = table_container.table_item
                page_items[table_container.indexs[0]] = table_item
                for idx in table_container.indexs[1:]:
                    table_item.replace_items.append(page_items[idx])
                    page_items[idx] = None
                page_items[table_container.table_idx] = None

        return [item for item in page_items if item is not None]

    def run(self, items: List[AtomItem]) -> List[AtomItem]:
        rtn = []
        page_grouped_items = []
        for _, group in itertools.groupby(items, key=lambda x: x.start_page):
            page_grouped_items.append(list(group))
        for group in page_grouped_items:
            table_dealed_group = self.deal_table(group)
            rtn.extend(table_dealed_group)
        return rtn


class TableQualityRepair(BaseModel):
    """
        提高表格质量
        1) 去除空列数据
        2) 移除不符合要求的表格
    """
    COL_MIN_NUM: int = 2
    """表格最小列数量"""
    MAX_HEADER_CELL_LEN: int = 30
    UNNAME_HEADER_PREXI: str = "unname_"

    def _empty_header_deal(self, df: DataFrame) -> DataFrame:
        cols = df.columns.tolist()
        new_columns = []
        for idx, col in enumerate(cols):
            if col:
                new_columns.append(col)
            else:
                new_columns.append(self.UNNAME_HEADER_PREXI + str(idx))
        df.columns = new_columns
        return df

    def remove_empty_col(self, table: TextPDFTableAtomItem):
        """移除数据都为空的列"""
        df = self._empty_header_deal(table.df)

        def check_empty(data):
            counter = Counter([x for x in data if not x])
            empty_num = counter[None] + counter['']
            return empty_num == len(data)

        droped_cols = df.columns[df.apply(check_empty)]
        table.df = df.drop(columns=droped_cols)

    def is_validated_table(self, table: TextPDFTableAtomItem) -> bool:
        headers = table.df.columns.tolist()
        flag = all(len(header) < self.MAX_HEADER_CELL_LEN for header in headers)
        flag2 = len(headers) >= self.COL_MIN_NUM
        return flag and flag2

    def run(self, items: List[AtomItem]) -> List[AtomItem]:
        rtn = []
        for item in items:
            if item.item_type == TAtomItem.TABLE:
                table: TextPDFTableAtomItem = item
                self.remove_empty_col(table)
                if self.is_validated_table(table):
                    rtn.append(item)
                else:
                    rtn.extend(table.replace_items)
            else:
                rtn.append(item)
        not_empty_rtn =  [item for item in rtn if item]
        return not_empty_rtn


class TextPdfRepair(BaseModel):
    cross_page_table_repair: CrossPageTableRepair = CrossPageTableRepair()
    table_desc_repair: TableDescRepair = TableDescRepair()
    image_desc_repair: ImageDescRepair = ImageDescRepair()
    table_pos_repair: TablePosRepair = TablePosRepair()
    image_pos_repair: ImagePosRepair = ImagePosRepair()
    table_quality_repair: TableQualityRepair = TableQualityRepair()

    @staticmethod
    def adjust_table(table_item: TableAtomItem):
        # 删除数据为空字符串|None的列
        df = table_item.df

        def check_empty(data):
            counter = Counter([x for x in data if not x])
            empty_num = counter[None] + counter['']
            return empty_num == len(data)

        table_item.df = df.drop(columns=df.columns[df.apply(check_empty)])

    def run(self, items: List[AtomItem]) -> List[AtomItem]:
        items = self.image_pos_repair.run(items)
        items = self.table_pos_repair.run(items)
        items = self.cross_page_table_repair.run(items)
        items = self.table_quality_repair.run(items)
        items = self.table_desc_repair.run(items)
        items = self.image_desc_repair.run(items)
        return items


class TextPdfTitleRepair(BaseModel):
    tconn: TitleConnector

    def run(self, titles: List[TitleNode], items: List[AtomItem]) -> List[AtomItem]:
        """
        标题链接到item上
        """
        has_index_items: List[tuple[AtomItem, int]] = []
        for index, item in enumerate(items):
            has_index_items.append((item, index))
        title_conn_pos_dict: Dict[int, TitleNode] = {}
        cur_find_pos = 0
        for title in titles:
            candcates = has_index_items[cur_find_pos:]
            new_pos = self.tconn.find_similar_title(title, candcates)
            if new_pos < 0:
                continue
            title_conn_pos_dict[new_pos] = title
            cur_find_pos = new_pos
        # 替换处理
        final_items = []
        for index, item in enumerate(items):
            if index in title_conn_pos_dict:
                title = title_conn_pos_dict[index]
                title_item = TitleAtomItem.instance(title.title, title.title_level, item.start_page)
                final_items.append(title_item)
            else:
                final_items.append(item)
        return final_items



from abc import abstractmethod, ABC
from typing import List

import pandas as pd
from pandas import DataFrame
from pptx.shapes.autoshape import Shape
from pptx.shapes.graphfrm import GraphicFrame
from pptx.slide import Slide
from pydantic import BaseModel

from pfip.base.errors import UnImplementException
from pfip.base.parser.parser_model import AtomItem, TextAtomItem, TableAtomItem, TitleAtomItem


class PPTSlideHandler(ABC, BaseModel):
    @abstractmethod
    def support(self, slide: Slide) -> bool:
        raise UnImplementException()

    @abstractmethod
    def run(self, slide: Slide, page_num: int) -> List[AtomItem]:
        raise UnImplementException()


class PPTShapeHandler(ABC, BaseModel):
    @abstractmethod
    def support(self, shape: Shape) -> bool:
        raise UnImplementException()

    @abstractmethod
    def run(self, shape: Shape, page_num: int) -> List[AtomItem]:
        raise UnImplementException()


class PPTParagraphHandler(PPTShapeHandler):
    def support(self, shape: Shape) -> bool:
        return shape.has_text_frame

    def run(self, shape: Shape, page_num: int) -> List[TextAtomItem]:
        rtn = []
        for paragraph in shape.text_frame.paragraphs:
            for run in paragraph.runs:
                rtn.append(TextAtomItem(content=run.text, start_page=page_num, end_page=page_num))
        return rtn


class PPTTableHandler(PPTShapeHandler):
    def support(self, shape: Shape) -> bool:
        return shape.has_table

    @staticmethod
    def to_data_frame(shape: GraphicFrame) -> DataFrame:
        table = shape.table
        data = []
        columns = []
        for i, row in enumerate(table.rows):
            if i == 0:
                columns.extend([cell.text for cell in row.cells])
            else:
                data.append([cell.text for cell in row.cells])
        return pd.DataFrame(data, columns=columns)

    def run(self, shape: GraphicFrame, page_num: int) -> List[TableAtomItem]:
        df = self.to_data_frame(shape)
        table_item = TableAtomItem.simple(df=df, page_num=page_num)
        return [table_item]


class PPTTitleHandler(PPTSlideHandler):
    def support(self, slide: Slide) -> bool:
        return slide.shapes.title is not None

    def run(self, slide: Slide, page_num: int) -> List[TitleAtomItem]:
        slide_title = slide.shapes.title.text
        title_item = TitleAtomItem(
            content=slide_title,
            start_page=page_num,
            end_page=page_num,
            title_level=1,
        )
        return [title_item]


class PPTNoteHandler(PPTSlideHandler):
    def support(self, slide: Slide) -> bool:
        return slide.has_notes_slide

    def run(self, slide: Slide, page_num: int) -> List[TextAtomItem]:
        text = slide.notes_slide.notes_text_frame.text
        node_item = TextAtomItem(content=text, start_page=page_num, end_page=page_num)
        return [node_item]
